#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include "PS2Defines.h"
#include "AddRegisters.h"
#include "texture.h"
#include <math.h>

/*

Third example based on the tutorial from Henry S Fortuna...

This is just one example that uploads a background to the GS.

Read the eight first lessons from Henry S Fortuna. (http://www.hsfortuna.pwp.blueyonder.co.uk/)
The last lesson you read should be the Using DMAC Tags lesson.

This is basically an excersice in DMAC Tags...
In order to send more than 4096 bytes of data in one DMAC transfer we need to use DMAC Tags.
What we want to do is to send a 640x480 background picture to the GS to display in the background.
There is a limitation of 512K in the amount of data you can load up to the GS per IMAGE GIF tag.
Since our background is 1.2 MB we need to divide the background into three GIF sections.
There are two problems with this though, all our DMA tags will not fit into 4096 bytes of memory and 
the memory allocated for the DMA sections may not be continous.
So we will allocate 8192 bytes for our DMA tags, the first 4096 bytes will have the first two GIF sections
and the last 4096 bytes for the last GIF section.

This is how our Tags will look like in memory:

-=*=-=*=-=*=- First 4096 bytes in the 8192 bytes for DMAC Tags -=*=-=*=-=*=-
Tag1: DMA_ID_SOURCE_CNT, QWC = 6
===
6 Qwords that specifies our first GIF Section.
We tell the GS that we want to upload 640x160 pixels.
===
Tag2: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(first 4096 bytes of the first 640x160 pixel block).
Tag3: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(second 4096 bytes of the first 640x160 pixel block).
Tag4: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(third 4096 bytes of the first 640x160 pixel block).
.
.
Tag101: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(last 4096 bytes of the first 640x160 pixel block).

OK, we still got room in the first 4096 bytes for DMAC Tags, let's add the next section...

Tag102: DMA_ID_SOURCE_CNT, QWC = 6
===
6 Qwords that specifies our second GIF Section.
We tell the GS that we want to upload 640x160 pixels.
===
Tag103: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(first 4096 bytes of the second 640x160 pixel block).
Tag104: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(second 4096 bytes of the second 640x160 pixel block).
Tag105: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(third 4096 bytes of the second 640x160 pixel block).
.
.
Tag202: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(last 4096 bytes of the second 640x160 pixel block).

Here is where we would run into problems if we just continue add tags, we cannot be sure that the 8192 bytes are continuous
so we need to add a DMA_ID_SOURCE_NEXT tag that points to the last 4096 bytes for the DMAC tags.
Note that this tag and the 6 consequtive QWords for the last GIF sections is still located in the first 4096 bytes for the DMAC tags...

Tag203: DMA_ID_SOURCE_NEXT, QWC = 6, ADDR = sps2GetPhysicalAddress(Second 4096 bytes fo the 8192 DMA Tags block)
===
6 Qwords that specifies our third GIF Section.
We tell the GS that we want to upload 640x160 pixels.
===

-=*=-=*=-=*=- Second 4096 bytes in the 8192 bytes for DMAC Tags -=*=-=*=-=*=-
Tag204: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(first 4096 bytes of the third 640x160 pixel block).
Tag205: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(second 4096 bytes of the third 640x160 pixel block).
Tag206: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(third 4096 bytes of the third 640x160 pixel block).
.
.
Tag303: DMA_ID_SOURCE_REF, QWC = 256, ADDR = sps2GetPhysicalAddress(last 4096 bytes of the third 640x160 pixel block).

Don't forget to end the transfer.

Tag304: DMA_ID_SOURCE_END, QWC = 0

One new thing to consider is that the allocated memory is cached (SPS2_MAP_CACHED) so we need to call sps2FlushCache(iSPS2Desc)
before we fire off the DMAC tags. This is to ensure the DMAC tags really are written to the correct memory location.
To fire them off, just point the DMAC to the first tag and use CHCR_MOD_CHAIN in stead of CHCR_MOD_NORMAL.

Written by Bamse 2004...
bamse@playstation2-linux.com

*/

// These are made global so we can access them from anywhere
sps2Memory_t * pMemCached;
int iSPS2Desc;
Dn_CHCR_t chcr;	  			// Channel control register used to control the DMA transfer.
Dn_MADR_t madr;
char * pMemPosition;

// Work out how we are going to allocate SPS2 to each use:
// Here we have 1200K allocated for the 640x480 background, that is right at the start of the memory
// Next, 8K allocated for the texture upload function to use for DMAing
// (The KILOBYTES macro just multiplies the number by 1024)
#define TEXTUREMEM_1200K	0
#define TEXTUREDMAMEM_8K	(KILOBYTES(1200))
#define TOTALMEMNEEDED		(KILOBYTES(1208))

int main(void)
{
	//initialise sps2
	iSPS2Desc = sps2Init();

	// Initialise the screen because we are drawing graphics
	sps2UScreenInit(0);

	// Allocate the SPS2 memory
	pMemCached = sps2Allocate(TOTALMEMNEEDED, 
							SPS2_MAP_BLOCK_4K |	SPS2_MAP_CACHED, 
							iSPS2Desc);

	// Get a char pointer, we can use this to access memory addresses
	// relative to the start of SPS2 memory
	char * pcMemory = (char *)pMemCached->pvStart;

	// Load the 640*480 bitmap into SPS2 memory
	// The final parameter says whether or not to make RGB(0,0,0) transparent
	if(!LoadBitmap("Background.bmp", &pcMemory[TEXTUREMEM_1200K], 640, 480, false))
	{
		printf("Failed to load Background texture.");
		exit(0);
	}

	while(true)
	{
		// Clear screen.
		sps2UScreenClear(0x00, 0x00, 0x00);

		// Upload the Background, first pointer is to the picture and the second one is to the memory where we insert our DMA and GIF tags.
		UploadBackground(sps2UScreenGetDrawBuffCurrent(), &pcMemory[TEXTUREMEM_1200K], &pcMemory[TEXTUREDMAMEM_8K]);

 		// Display the Background
		sps2UScreenSwap();
	}
	
	// Shutdown the screen
	sps2UScreenShutdown();

	// Release SPS2
	sps2Release(iSPS2Desc);

	return 0;
}
